#!/usr/bin/python
#Dependencies: sox libsox-fmt-all octave curl
#Additionally, if you're on python2.6: python-argparse python-imaging

import sys, subprocess, time, os, argparse, logging, random, operator
import solver, spectrogram_splitter, simple_mash, contextual_mash
from parse_solver_args import parse_solver_options,print_solver_help_and_exit

mplayer_process = None
def play_file(filename):
    global mplayer_process
    if mplayer_process != None: #Don't play two audio files at once
        mplayer_process.kill()
    mplayer_process = subprocess.Popen(["mplayer", filename], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)


parser = argparse.ArgumentParser(description="Runs the complete captcha solving process with different values for the number of lowest captchas to find, in order to find the best value")

#Solver options
parser.add_argument('-s', nargs='*', action='append', help='Adds a solver to the list of solvers to be used to solve the captcha. ' 
                    + 'This option can be specified multiple times to use additional solvers as backup.  Use -sh for information on specifying solvers')
parser.add_argument('-sh', required=False, action='store_true', help='Show the syntax to specify solvers')

#Splitting args
parser.add_argument('-b', type=int, required=False, default=140, metavar='black_pixel_count_threshold', help='The maximum number of black pixels a column in the spectrogram can have and still be considered part of a word')

#Misc args
parser.add_argument('-a', required=False, action='store_true', help='Save debug images and audio files to help with debugging the splitter')
parser.add_argument('-d', required=False, action='store_true', help='Don\'t mash words')
parser.add_argument('-f', required=False, action='store_true', help='Only submit five words so we can be sure that we get those ones right (Implies -d)')
parser.add_argument('-i', required=False, action='store_true', help='Interactive mode: Ask for confirmation about each captcha')
parser.add_argument('-ifile', required=False, default='interactive.log', metavar="interactive.log", help="In interactive mode, the file to log captchas that are changed by the user.")
parser.add_argument('-l', type=int, required=False, default=sys.maxint, metavar='limit', help='The number of captchas to solve')
parser.add_argument('-m', required=False, action='store_true', help='Display timing information about the captchas solved')
parser.add_argument('-n', required=False, default=-1, type=int, metavar='number', help='Save the files used for solving captchas.  A number to start at when naming the captchas should be supplied as well.  They will be named according to the scheme: current####.')
parser.add_argument('-sm', required=False, action='store_true', help='Use the simple masher (as opposed to the default of the contextual masher)')
parser.add_argument('-w', type=int, required=False, default=0, metavar='seconds', help='The number of seconds to wait between rounds')
parser.add_argument('-v', required=False, action='store_true', help='Verbose mode')
parser.add_argument('-vv', required=False, action='store_true', help='Very Verbose mode')

args = parser.parse_args()
if args.sh:
    print_solver_help_and_exit()

output_level = logging.CRITICAL
if args.vv:
    output_level = logging.DEBUG
elif args.v:
    output_level = logging.INFO

if args.f:
    args.d = True

masher = None
if not args.d:
    if args.sm:
        masher = simple_mash.SimpleMash()
    else:
        masher = contextual_mash.ContextualMash(output_level)
splitter = spectrogram_splitter.SpectrogramSplitter(black_pixel_count_threshold = args.b, debug = output_level, debug_images = args.a)
solver_list, solver_thresholds = parse_solver_options(args.s, output_level)
captcha_solver = solver.CaptchaSolver(solver_list, splitter, solver_thresholds, masher, output_level)

save_files = (args.n != -1)
if save_files:
    captcha_solver.make_folders()
    start = args.n
else:
    start = os.getpid()

if args.i:
    interactiveFile = open(args.ifile, "a")

correct = 0
bad_words = { }
streak = 0
combo = 0
i = start
while i <  start + args.l:
    if save_files:
        filename = "current" + str(i) 
    else:
        filename = "current" + str(start)

    challenge = ""

    startTime = time.time()
    challenge, url = captcha_solver.get_audio_captcha_url()
    if not captcha_solver.get_mp3(url, filename):
        print "Could not download mp3 file",filename + ".mp3"
        continue
    if args.i:
        play_file(captcha_solver.tempdir + filename + ".mp3")
    doneDownloadingTime = time.time()

    #Solve the captcha word by word
    answer = captcha_solver.solve(filename)
    doneSolveTime = time.time()

    deleteFiles = False
    before_user = answer
    if args.i:
        userInput = ','
        while userInput == ',':
            userInput = raw_input("Did this just say '" + answer + "'? (',' for to replay, 'd' to delete these files, enter for yes, or type the real answer)")
            userInput = userInput.strip()
            if userInput == ',':
                play_file("samples/" + filename + ".mp3")
        if userInput == "d":
            deleteFiles = True
        elif userInput != "":
            answer = userInput

    answer_parts = answer.split(" ")
    before_user_parts = before_user.split(" ")
    for j in range(0, len(answer_parts)):
        if answer_parts[j] == "." and j < len(before_user_parts):
            answer_parts[j] = before_user_parts[j]
    for j in range(len(answer_parts), 6):
        if j < len(before_user_parts): #Make sure to check in case we only find 5 words
            answer_parts.append(before_user_parts[j])
    answer = " ".join(answer_parts)

    if before_user != answer:
        interactiveFile.write(filename + "," + before_user + "," + answer + "\n")

    if args.f and len(answer_parts) > 5:
        answer = " ".join(answer_parts[0:5])

    if answer == "":
        print "No answer found for",filename
        success = False
    else:
        print "Submitting " + answer 
        response = captcha_solver.respond_to_captcha(challenge, answer, "Button1=Submit") #Tell google the answer
        success = response.find("Correct!") != -1
        doneSubmittingTime = time.time()
        captcha_solver.feedback(success, answer, filename)
        if deleteFiles or not save_files:
            captcha_solver.delete_files(success, filename)
        else:
            captcha_solver.save_files(success, answer, filename)  #Save all the resulting files

        if success:
            correct += 1
        
        sys.stdout.flush()
        endTime = time.time()
        if args.m:
            print "Time taken to get challenge and download: ", doneDownloadingTime - startTime
            print "Time taken to solve (and split if necessary) the captcha: ", doneSolveTime - doneDownloadingTime
            print "Time taken to submit the captcha: ", doneSubmittingTime - doneSolveTime
            print "Total time: ", doneSubmittingTime - startTime

    if success:
        print "Correct ",
        streak += 1
        print "(" + str(streak) + " in a row)",
        if streak > combo:
            combo = streak
    else:
        print "Incorrect ",
        streak = 0

        if args.f:
            for word in answer.split(" "):
                if word not in bad_words:
                    bad_words[word] = 0
                bad_words[word] += 1
            print "Word miss dictionary:", bad_words
    print "(Highscore " + str(combo) + ")",
    print correct,"/",i - start + 1," = ",round(float(correct)/float(i - start + 1) * 100, 2),"%"

    time.sleep(args.w)
    i += 1

if args.i:
    interactiveFile.close()

