import hashlib, logging, time, os, argparse, sys, struct, operator
from solverbase import SolverBase

class Md5Solver(SolverBase):
    """This class will attempt to solve a captcha by looking it up in the md5sum file"""

    classname = 'Md5Solver'
    accuracy_multiplier = 100 #If md5solver solves it, always choose this answer
    split_components = [ ]
    incorrect_matrix = { 0 : { "" : -sys.maxint }, 1 : { "" : -sys.maxint }, 2 : { "" : -sys.maxint }, 3 : { "" : -sys.maxint }, 4 : { "" : -sys.maxint }, 5 : { "" : -sys.maxint } }
    def __init__(self, md5_file = 'md5sums.small', word_map_file = 'md5sums.map', debug = logging.CRITICAL, filename_prefix = "temp/", 
                    filename_suffix = ".mp3", basepath = ''):
        """ Loads the word mapping file and initializes the Md5 solver
            Parameters: 
                md5_file = the file with all of the md5s in them.  This binary file should have fixed size rows each 28 bytes
                            long.  Each row should be: <md5sum of the captcha (16 bytes)><words 0-5 (2 bytes each)>
                            Where each word is just an index into the matching word_map_file
                word_map_file = the file that can be used to look up word indexes in from the md5_file.  Each word 
                            should be on its own line in the format: word,index
                debug = how much information should be printed
                filename_prefix = A filename prefix to add to the beginning of all filenames passed in
                filename_suffix = A filename suffix to add to the end of all the filenames passed in
        """
        if basepath == "":
            self.basepath = os.path.dirname(__file__) + '/'
        else:
            self.basepath = basepath
        self.logger = logging.getLogger(self.classname)
        logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger.setLevel(debug)

        self.filename_prefix = filename_prefix
        self.filename_suffix = filename_suffix

        self.md5file = open(self.basepath + md5_file, "rb")
        self.num_md5s = os.path.getsize(self.basepath + md5_file) / 28
        self.word_map = {}
        wordfile = open(self.basepath + word_map_file, "r")
        lines = wordfile.readlines()
        for line in lines:
            parts = line.replace("\n","").split(",")
            self.word_map[parts[1]] = parts[0]
        wordfile.close()

        self.logger.debug("Md5 file has %d entries", self.num_md5s)

    def finish_solving(self,filename):
        """Attempts to find an md5 hash in the list of solved captcha md5s.  Returns the answer if found, otherwise returns an empty string"""
        md5 = self.get_md5(self.filename_prefix + filename + self.filename_suffix)
        firstHalf = int(md5[0:16],16)
        secondHalf = int(md5[16:32],16) #Stupid array won't take a 16 byte long
        answer = self.binarySearch((firstHalf, secondHalf), 0, self.num_md5s - 1)
        if answer != "":
            answer_matrix = {}
            for i in range(0, len(answer)):
                answer_matrix[i] = {self.word_map[str(answer[i])] : sys.maxint}
            return answer_matrix
        return self.incorrect_matrix #Return an empty matrix because we couldn't find the answer

    def binarySearch(self, searchValue, left, right):
        """Performs a binary search (Adapted from: http://en.wikibooks.org/wiki/Algorithm_Implementation/Search/Binary_search)"""
        if right < left:
            return ""
        mid = (left + right) / 2
        mid_value, answer = self.value_at_index(mid)
        if searchValue[0] > mid_value[0]:
            return self.binarySearch(searchValue, mid + 1, right)
        elif searchValue[0] < mid_value[0]:
            return self.binarySearch(searchValue, left, mid - 1)
        else:
            if searchValue[1] > mid_value[1]:
                return self.binarySearch(searchValue, mid + 1, right)
            if searchValue[1] < mid_value[1]:
                return self.binarySearch(searchValue, left, mid - 1)
            return answer

    def value_at_index(self, index):
        """Gets the md5 value and answer at the specified index in the md5file"""
        if index > self.num_md5s or index < 0:
            return ""
        else:
            self.md5file.seek(index * 28)
            binary_data = self.md5file.read(28)
            firstHalf, secondHalf, answer1,answer2,answer3,answer4,answer5,answer6 = struct.unpack("QQHHHHHH", binary_data)
            answer = [answer1,answer2,answer3,answer4,answer5,answer6]
            return (firstHalf, secondHalf), answer

    def get_md5(self,filename):
        """Gets the md5 hash of the file passed in"""
        m = hashlib.md5()
        captchaFile = open(filename, "rb")
        m.update(captchaFile.read())
        captchaFile.close()
        return m.hexdigest()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Looks up all the md5s of the *.mp3 files in the given folder, to find out if we've already solved them")
    parser.add_argument('dir', type=str, help='The directory to find *.mp3 files in')
    parser.add_argument('-f', type=str, default="md5sums.small", help='The md5 dictionary file')
    parser.add_argument('-v', required=False, action='store_true', help='Verbose mode.')
    parser.add_argument('-vv', required=False, action='store_true', help='Very Verbose mode.')
    parser.add_argument('-w', type=str, default='md5sums.map', help='The word mapping file')
    args = parser.parse_args()    

    debug_level = logging.INFO
    if args.v or args.vv:
        debug_level = logging.DEBUG
    files = os.listdir(args.dir)

    md5_solver = Md5Solver(debug = debug_level, md5_file = args.f, word_map_file = args.w, filename_suffix="", filename_prefix="")
    startTime = time.time()
    correct = 0
    for filename in files:
        if filename.endswith(".mp3"):
            md5_solver.start_solving(args.dir + "/" + filename)
            answer = md5_solver.finish_solving(args.dir + "/" + filename)
            if answer != md5_solver.incorrect_matrix:
                correct += 1
                if args.vv:
                    print filename,"has an answer of",
                    for answer_matrix in answer.items():
                        print max(answer_matrix[1].iteritems(), key=operator.itemgetter(1))[0],
                    print ""

    print "Looked up",len(files),"captchas in ",time.time()-startTime,"seconds.  Found", correct


