import logging, time, subprocess, argparse, operator, math, sys, os
from solverbase import SolverBase

class PhashSolver(SolverBase):
    """This class will attempt to solve a captcha by using an image comparison (via phash) on the spectrogram"""

    classname = "PhashSolver"
    accuracy_multiplier = .5 #It's only half as accurate as the neural network 
    split_components = ["image"]
    def __init__(self, debug = logging.CRITICAL, hashfile = "hashes140", number_to_find = 5, filename_prefix = "temp/", filename_suffix = ".png", basepath = ""):
        """Initializes a solver that uses pHash to solve the captcha
            Parameters:
                debug = how much information should be printed
                hashfile = The file with all of the pHashes in it for the sample set in it
                number_to_find = the number of captchas for each word in the sample set
                filename_prefix = A filename prefix to add to the beginning of all filenames passed in
                filename_suffix = A filename suffix to add to the end of all the filenames passed in
        """
        if basepath == "":
            self.basepath = os.path.dirname(__file__) + '/'
        else:
            self.basepath = basepath
        self.logger = logging.getLogger(self.classname)
        logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger.setLevel(debug)
        
        self.hashfile = hashfile
        self.number_to_find = number_to_find
        self.filename_prefix = filename_prefix
        self.filename_suffix = filename_suffix
        self.reset_solver()

        if sys.maxsize > 2**32:
            self.programname = "find_cluster64"
        else:
            self.programname = "find_cluster32"

        self.deteriate_factor = 150 #How fast the accuracy deteriates once your distance is larger than the minimum one

    def reset_solver(self):
        """Resets the state of the phash solver"""
        self.processes = []

    def start_solving(self, filename):
        """Starts solving an entire captcha using phash"""
        self.reset_solver()
        for i in range(0, 6): #Start the find_cluster processes
            self.start_solving_word(self.filename_prefix + filename + "_" + str(i) + self.filename_suffix)

    def finish_solving(self, filename):
        """Finishes solving the entire captcha using phash"""
        predMatrices = {}
        for i in range(0, len(self.processes)):  #Wait for the find_cluster processes to end and get the results
            predMatrices[i] = self.finish_solving_word(i)
        return predMatrices

    def start_solving_word(self, filename):
        """Starts solving a single word of a captcha using phash, by calling the find_cluster program"""
        options = [ self.basepath + self.programname, filename , "-f", self.basepath + self.hashfile, "-l", str(self.number_to_find), "-v"]
        self.processes.append(subprocess.Popen(options, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE))

    def finish_solving_word(self, index):
        """Finishes solving a single word of a captcha using phash, by parsing the results of find_cluster"""
        output = self.processes[index].communicate()[0]
        lines = output.split("\n")
        matrix = {}
        for line in lines:
            if line != "":
                parts = line.replace("\n","").split(":")
                matrix[parts[1]] = 1 - float(parts[0])/(40.0 * self.number_to_find) #The maximum distance between two images 28
                if matrix[parts[1]] < 0:
                  print "Got distance greater than what was expected, adjust the maximum distance per word"
                  print parts, matrix[parts[1]]
                  sys.exit()
        
        distanceFreq = {}
        for distance in matrix.values():
            if distance not in distanceFreq:
                distanceFreq[distance] = 1
            else:
                distanceFreq[distance] += 1
        distanceFreq = sorted(distanceFreq.iteritems(), key=operator.itemgetter(0))

        likelihood = {}
        multiplier = 100
        for i in range(len(distanceFreq)):
            distance,count = distanceFreq[i]
            if i != 0:
                multiplier *= self.deteriate_factor * (distance - distanceFreq[i-1][0]) 
            for word, percent in matrix.items():
                if distance == percent:
                    likelihood[word] = percent * multiplier
            
        total = 0
        for word, percent in likelihood.items():
            total += percent
        for word, percent in likelihood.items():
            likelihood[word] = percent / total
        
        return likelihood

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Solves a single word of a captcha using the phash method.")
    parser.add_argument('-f', nargs='*', required=True, type=str, action='append', help='The png file(s) of the single word to solve')
    parser.add_argument("-a", required=False, type=str, default="hashes140", help='The hash file to feed into find_cluster')
    parser.add_argument("-n", required=False, type=int, default=5, help='The number of similar words to find when using find_cluster')
    parser.add_argument('-v', required=False, action='store_true', help='Verbose mode.')
    parser.add_argument('-vv', required=False, action='store_true', help='Very Verbose mode.')
    args = parser.parse_args()    

    debug_level = logging.INFO
    if args.v or args.vv:
        debug_level = logging.DEBUG

    phash_solver = PhashSolver(debug = debug_level, hashfile = args.a, number_to_find = args.n)
    for filenames in args.f:
      for filename in filenames:
        startTime = time.time()

        phash_solver.reset_solver()
        phash_solver.start_solving_word(filename)
        matrix = phash_solver.finish_solving_word(0)
        answer = max(matrix.iteritems(), key=operator.itemgetter(1))[0]

        if args.v or args.vv:
            print "Phash determined",filename,"has an answer of",answer,"in",time.time()-startTime,"seconds."
        else:
            print filename + "," + answer

