import subprocess, os, shutil, operator, logging, hashlib

class CaptchaSolver:
    """Wraps all the methods required to defeat recaptcha"""

    tempdir = "temp/"
    def __init__(self, solvers, splitter, solver_thresholds = [0], masher = None,
                debug = logging.CRITICAL, site_key = "6Ld4iQsAAAAAAM3nfX_K0vXaUudl2Gk0lpTF3REf", submit_url = "http://www.google.com/recaptcha/demo/"):
        """Declares a new CaptchaSolver class that will try to solve each word with the given solvers.  If a solver reaches the corresponding value
           for certainness in solver_thresholds, that solver's answer will be selected.  Otherwise it will continue to try other solvers.  The 
           captcha will be split using the given splitter, and all answers will be run through the given word masher if one was given.  The parameter
           debug determines how much information is printed.  site_key should be the recaptcha site_key for the site whom captchas you wish to break,
           and submit_url should be the url to submit the captcha answers to."""
        self.logger = logging.getLogger('CaptchaSolver')
        logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger.setLevel(debug)

        self.site_key = site_key
        self.submit_url = submit_url

        self.solver_thresholds = solver_thresholds
        self.solvers = solvers
        self.split_components = [] #Holds a list of all the components needed from splitting
        for solver in self.solvers:
            for split_component in solver.split_components:
                if split_component not in self.split_components:
                    self.split_components.append(split_component)

        self.splitter = splitter
        self.masher = masher
        if not os.path.exists(self.tempdir): #Create the temp folder
            os.mkdir(self.tempdir)


    def make_folders(self, root= "automated/"):
        """Make the necessary folders to save the temp files to (using save_files)"""
        if root[-1] != "/":
            root += "/"
        subfolders = [ "samples", "spectrograms", "solved_clusters", "incorrect/",
                        "incorrect/samples", "incorrect/spectrograms", "incorrect/solved_clusters" ]
        if not os.path.exists(root):
            os.mkdir(root)
        for folder in subfolders:
            if not os.path.exists(root + folder):
                os.mkdir(root + folder)

    def get_audio_captcha_url(self):
        """Gets the challenge token and corresponding url to download the audio captcha from"""
        url = "http://www.google.com/recaptcha/api/challenge?type=audio&k="
        beginString = "challenge : '"
        endString = "'"

        options = ["curl", "-s", url + self.site_key]
        curl = subprocess.Popen(options, stdin = subprocess.PIPE, stdout = subprocess.PIPE)
        output = curl.communicate()[0]

        start = output.find(beginString)
        start = start + len(beginString)
        end = output.find(endString, start + 1)
        challenge = output[start:end]
        return challenge, "http://www.google.com/recaptcha/api/image?c=" + challenge

    def get_mp3(self, url, filename):
        """Get the mp3 file at the corresponding url.  Save the file to the $tempdir/filename.mp3.  Returns whether
            or not the file is an mp3 file"""
        options = ["curl", "-s", url, "-o", self.tempdir + filename + ".mp3"]
        curl = subprocess.Popen(options, stdin = subprocess.PIPE, stdout = subprocess.PIPE)
        curl.wait()
        return self.is_mp3(self.tempdir + filename + ".mp3")

    def is_mp3(self, filename):
        """Returns whether or not filename is a mp3 file by running 'file' on it"""
        p = subprocess.Popen(["file", filename], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        p.wait()
        output = p.communicate()[0]
        if output.find("MPEG") == -1:
            self.logger.debug("Got " + output.replace("\n","") + " filetype from google")
            return False
        return True

    def solve(self, filename):
        """Solves an entire captcha"""
        answer = ""
        split = False
        start_solver = 0
        predMatrices = {}

        self.solvers[0].reset_solver()
        if len(self.solvers[0].split_components) != 0:
            self.splitter.split_captcha(filename, self.split_components)
            split = True
        self.solvers[0].start_solving(filename)
        predMatrices[0] = self.solvers[0].finish_solving(filename)

        #If md5s couldn't find it, just move on to the next solver now.  Rather than wait and move to the next one word at a time (saves overall time)
        if self.solvers[0].classname == "Md5Solver" and predMatrices[0] == self.solvers[0].incorrect_matrix and len(self.solvers) > 1: 
            start_solver += 1
            self.solvers[1].reset_solver()
            if len(self.solvers[1].split_components) != 0:
                self.splitter.split_captcha(filename, self.split_components)
                split = True
            self.solvers[1].start_solving(filename)
            predMatrices[1] = self.solvers[1].finish_solving(filename)

        for i in range(0, 6):
            word = ""
            pwords=[]
            foundThreshold = False

            for j in range(start_solver, len(self.solvers)):
                word = max(predMatrices[j][i].iteritems(), key=operator.itemgetter(1))[0]

                if word != "":
                    self.logger.info("%s (solver %d) choosing %s for word %d with %.4f", self.solvers[j].classname, j, word, i, predMatrices[j][i][word])
                else: #The Md5Solver can't always choose a word
                    self.logger.info("%s (solver %d) couldn't choose an answer for word %d", self.solvers[j].classname, j, i)
                  
                pwords.append([predMatrices[j][i][word] * self.solvers[j].accuracy_multiplier,word])

                if self.solver_thresholds[j] < predMatrices[j][i][word]: 
                    foundThreshold = True
                    break

                if j + 1 < len(self.solvers):
                    solver = self.solvers[j+1]
                    if len(solver.split_components) != 0 and not split:
                        self.splitter.split_captcha(filename, self.split_components)
                        split = True

                    solver.reset_solver()
                    solver.start_solving_word(solver.filename_prefix + filename + "_" + str(i) + solver.filename_suffix)
                    if j+1 not in predMatrices:
                        predMatrices[j+1] = {}
                    predMatrices[j+1][i] = solver.finish_solving_word(0)

            if not foundThreshold:
                certainty, word = max(pwords)
                
                best_solver = -1
                for solver_index in range(0, len(pwords)):
                    if certainty == pwords[solver_index][0] and word == pwords[solver_index][1]:
                        best_solver = solver_index
                        break
                if len(pwords) > 1:
                    self.logger.info("%s %d solved with the best match of %s with %f certainty", self.solvers[best_solver].classname,
                                    best_solver, word, certainty)
                best_pred_matrix = predMatrices[best_solver][i]
            else:
                self.logger.info("%s %d decided %s with certainity of %f (which is better than threshold %f)", self.solvers[j].classname, 
                                    j, word, predMatrices[j][i][word], self.solver_thresholds[j])
                best_pred_matrix = predMatrices[j][i]
            
            if answer != "":
                answer += " "

            if self.masher != None: 
                answer += self.masher.mash_word(word, best_pred_matrix)
            else:
                answer += word

        return answer

    def respond_to_captcha(self, challenge, answer, other_data):
        """Submit the captcha and any other data passed in to the submit_url specified when creating the class"""
        data = "recaptcha_challenge_field=" + challenge + "&recaptcha_response_field=" + answer + "&" + other_data

        options = ["curl", "-s", "-d", data, self.submit_url]
        curl = subprocess.Popen(options, stdin = subprocess.PIPE, stdout = subprocess.PIPE)
        output = curl.communicate()[0]

        return output

    def save_files(self, success, answer, filename, path = "automated/"):
        """Moves all the files from the captcha solving process to the correct directory"""
        if path[-1] != '/':
            path += "/"
        answer_parts = answer.split(" ")

        if not success:
           path += "incorrect/"

        #Save all the files accordingly
        shutil.move(self.tempdir + filename + ".mp3", path + "samples/" + filename + ".mp3")
        shutil.move(self.tempdir + filename + ".png", path + "spectrograms/" + filename + ".png")

        if os.path.exists(self.tempdir + filename + "_noborder.png"):
            shutil.move(self.tempdir + filename + "_noborder.png", path + "spectrograms/" + filename + "_noborder.png")
        if os.path.exists(self.tempdir + filename + "_debug.png"):
            shutil.move(self.tempdir + filename + "_debug.png", path + "spectrograms/" + filename + "_debug.png")
        for j in range(0,len(answer_parts)):
            splitfile = filename + "_" + str(j) + ".png"
            debugsplitfile = filename + "_" + str(j) + "before.png"
            wavsplitfile = filename + "_" + str(j) + ".wav"
            debugwavsplitfile = filename + "_" + str(j) + "before.wav"
            destdir = path + "solved_clusters/" + answer_parts[j]
            if not os.path.exists(destdir):
                os.mkdir(destdir)
            if os.path.exists(self.tempdir + splitfile):
                shutil.move(self.tempdir + splitfile, destdir + "/" + splitfile)
            if os.path.exists(self.tempdir + debugsplitfile):
                shutil.move(self.tempdir + debugsplitfile, destdir + "/" + debugsplitfile)
            if os.path.exists(self.tempdir + wavsplitfile):
                shutil.move(self.tempdir + wavsplitfile, destdir + "/" + wavsplitfile)
            if os.path.exists(self.tempdir + debugwavsplitfile):
                shutil.move(self.tempdir + debugwavsplitfile, destdir + "/" + debugwavsplitfile)

    def delete_files(self, success, filename):
        """Deletes all files downloaded / generated while solving the captcha"""
        files_to_delete = [self.tempdir + filename + ".mp3", self.tempdir + filename + ".png", 
                           self.tempdir + filename + "_noborder.png", self.tempdir + filename + "_debug.png"]
        for i in range(0,6):
            files_to_delete.append(self.tempdir + filename + "_" + str(i) + ".png")
            files_to_delete.append(self.tempdir + filename + "_" + str(i) + "before.png")
            files_to_delete.append(self.tempdir + filename + "_" + str(i) + ".wav")
            files_to_delete.append(self.tempdir + filename + "_" + str(i) + "before.wav")

        for delete_file in files_to_delete:
            if os.path.exists(delete_file):
                os.remove(delete_file)

    def feedback(self, success, mashed_answer, filename):
        """Records the MD5 sum if correct and lets the solvers know whether they got it right 
            so they can track internal accuracies"""
        if success:
            self.record_md5sum(self.tempdir + filename + ".mp3", mashed_answer)
        if len(self.solvers) == 1: #We really only care about tracking the solvers internal accuracy when using a single solver
            self.solvers[0].feedback(success)

    def record_md5sum(self, filename, mashed_answer):
        """Adds the md5 of the filename and answer to the new_md5s file"""
        m = hashlib.md5()
        captchaFile = open(filename, "rb")
        m.update(captchaFile.read())
        captchaFile.close()
        
        captchaFile = open("new_md5s", "a")
        captchaFile.write(m.hexdigest() + " " + mashed_answer.replace(" ", "_") + "\n")
        captchaFile.close()

