import Image, os, subprocess, operator, logging, argparse
from splitterbase import SplitterBase

class SpectrogramSplitter(SplitterBase):
    """Wraps all the methods required to split an audio recaptcha into the words using the spectrogram"""
    sample_rate = 11025.0
    black = (0, 0, 0)
    def __init__(self, black_pixel_count_threshold = 140, debug = logging.CRITICAL, debug_images = False, 
                    maximum_column_xdiff = 2, minimum_word_height = 170, filename_prefix = "temp/"):
        self.logger = logging.getLogger('SpectrogramSplitter')
        logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger.setLevel(debug)
        self.debug_images = debug_images

        self.maximum_column_xdiff = maximum_column_xdiff
        self.minimum_word_height = minimum_word_height
        self.black_pixel_count_threshold = black_pixel_count_threshold
        self.filename_prefix = filename_prefix
        self.image_index = {}

    def split_captcha(self, filename, split_components = [ "audio", "image" ]):
        """Splits a audio captcha and its spectrogram up into the indvidual components for each word.  If only an audio file or only a spectrogram
            image file is needed for each word, split_components can be used to only select the component you need."""
        self.get_spectrogram(filename)
        spectrogram = self.filename_prefix + filename + ".png"
        image = Image.open(spectrogram).convert("RGB")
        image = self.remove_white_border(image)
        self.image_index = {} #Reset the file nameing convention

        if self.debug_images:
            image.save(self.filename_prefix + filename + "_noborder.png", "PNG")
        images,column_xs = self.split_into_words(image, filename)

        self.logger.debug("Found %d images", len(images))

        if len(images) > 6:
            self.logger.warning("Found too many images, removing the thinnest ones")
            while len(images) > 6:
                new_column_xs = []
                new_images = []
                lengths = {}
                for i in range(0, len(column_xs)):
                    lengths[i] = column_xs[i][1] - column_xs[i][0]
                lengths = sorted(lengths.iteritems(), key=operator.itemgetter(1))
                lengths.reverse()

                indexes = []
                for i in range(0, 6):
                    indexes.append(lengths[i][0])
                indexes.sort()
                for i in indexes:                
                    new_images.append(images[i])
                    new_column_xs.append(column_xs[i])

                self.logger.debug("before: " + str(column_xs))
                self.logger.debug("lengths: " + str(lengths))
                self.logger.debug("indexes: " + str(indexes))
                self.logger.debug("after: " + str(new_column_xs))

                column_xs = new_column_xs
                images = new_images
        elif len(images) < 6:
            minimum_word_height = self.minimum_word_height
            while len(images) < 6:
                self.logger.warning("Found too few images, splitting the widest ones")
                minimum_word_height -= 5
                images,column_xs = self.split_into_words(image, filename, minimum_word_height)

        sizex,sizey = image.size
        if self.debug_images:
            self.split_wav_file(column_xs, sizex, filename, "before")

        for i in range(0, len(images)):
            split_image = images[i]
            if self.debug_images:
                self.save_image(split_image, filename, "before")
            split_image, (left, right) = self.trim_image(split_image)
            column_xs[i] = [column_xs[i][0] + left, column_xs[i][0] + right]
            if "image" in split_components:
                self.save_image(split_image, filename)
        if "audio" in split_components:
            self.split_wav_file(column_xs, sizex, filename)
        return len(images)

    def gc(self, x, y, sizex, sizey):
        """Returns whether or not the coordinates are in the image"""
        return x >= 0 and y >= 0 and x < sizex and y < sizey

    def save_image(self, image, filename, suffix = ""):
        if suffix not in self.image_index:
            self.image_index[suffix] = 0
        image.save(self.filename_prefix + filename + "_" + str(self.image_index[suffix]) +  suffix + ".png", "PNG")
        self.image_index[suffix] += 1

    def split_into_words(self, image, filename = "", minimum_word_height = -1, maximum_column_xdiff = -1):
        if maximum_column_xdiff == -1:
            maximum_column_xdiff = self.maximum_column_xdiff
        if minimum_word_height == -1:
            minimum_word_height = self.minimum_word_height

        sizex, sizey = image.size
        pixels = image.load()
        column_xs = []
        directions = [ -1, 1 ]
        x = 0

        while x < sizex:
            if not self.within_distance(pixels[x,minimum_word_height], self.black):
                column_x = [x,x]
                for i in range(0, len(directions)):
                    changed = True
                    while changed:
                        changed = False
                        for nx in range(column_x[i] + maximum_column_xdiff * directions[i], column_x[i], -directions[i]):
                            for ny in range(0, minimum_word_height):
                                if self.gc(nx,ny,sizex,sizey) and not self.within_distance(pixels[nx,ny], self.black):
                                    column_x[i] = nx
                                    changed = True
                                    break
                            if changed:
                                break
                if len(column_xs) != 0 and column_x[0] <= column_xs[-1][1]:
                    self.logger.debug("Extending word image %d from %d to %d", len(column_xs), column_xs[-1][1], column_x[1])
                    column_xs[-1][1] = column_x[1]
                else:
                    column_xs.append(column_x)
                x = column_x[1]
            x += 1

        self.logger.debug("Column x-coordinates are: %s", str(column_xs))

        cropped_images = []
        for (x1, x2) in column_xs:
            cropped_images.append(image.crop((x1 - 1, 0, x2 + 1, sizey)))

        if self.debug_images and filename != "":
            image = image.copy() #Get a copy so we don't mess up the cropped images
            pixels = image.load()
            for (x1, x2) in column_xs:
                for y in range(0, sizey):
                    pixels[x1,y] = (255,255,255)
                    pixels[x2,y] = (255,255,255)

            name = filename
            if name != "":
                name += "_"
            image.save(self.filename_prefix + name + "debug.png", "PNG")
        return cropped_images, column_xs

    def trim_image(self, image):
        sizex, sizey = image.size
        pixels = image.load()
        column_xs = []
        directions = [ -1, 1 ]

        left = 0
        while left < sizex:
            black_pixel_count = 0
            for y in range(0, self.minimum_word_height): #Count the number of black pixels in this column
                if self.within_distance(pixels[left,y], self.black):
                    black_pixel_count += 1
            if black_pixel_count < self.black_pixel_count_threshold:
                break
            left += 1

        right = sizex - 1
        while right >= 0:
            black_pixel_count = 0
            for y in range(0, self.minimum_word_height): #Count the number of black pixels in this column
                if self.within_distance(pixels[right,y], self.black):
                    black_pixel_count += 1
            if black_pixel_count < self.black_pixel_count_threshold:
                break
            right -= 1
        right += 1

        if left <= right: #If we've traversed the entire image, we've obviously failed.  Just don't crop
            self.logger.debug("Cropping image from (0,%d) to (%d,%d)",sizex, left, right)
            image = image.crop((left, 0, right, sizey))
        else:
            self.logger.debug("Could not trim image correctly, got borders as (%d,%d)", left, right)
            left = 0
            right = sizex
    
        return image, (left, right)

    def trim_wav(self, filepath, column_x, sizex):
        """Takes a wav file and trimms it to correspond a more narrow word column.  The file name and path should be passed in
             filepath, the start x and end x of the word's spectrogram should be passed as column_x, and the width of the wav's
             spectrogram should be passed in sizex"""
        soundsizex = self.get_duration(filepath)
        tempfile = filepath + "_temp.wav"

        skip = column_x[0] * float(soundsizex / sizex) * (1 / self.sample_rate) #How much to skip in seconds
        length = float(column_x[1] - column_x[0] + 1.0) * float(soundsizex / sizex) * (1 / self.sample_rate) #How long the wav should be in seconds

        self.logger.info("Skipping %f seconds and grabbing %f seconds. Size was %f pixels/%f frames and new size is %f pixels/%f frames",
                    skip, length, sizex, soundsizex, column_x[1] - column_x[0], (column_x[1] - column_x[0] + 1) * float(soundsizex / sizex))

        #Sox has the ability to split multiple parts at once, but I couldn't get it to work with 6 splits so we have to do it 6 times
        options = ["sox", filepath, tempfile, "trim", str(skip), str(length)] 
        process = subprocess.Popen(options, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        process.communicate() #wait for sox to finish
        os.rename(tempfile,filepath) #Sox can't have the input and output file with the same name

if __name__ == "__main__":    
    parser = argparse.ArgumentParser(description="Split an audio captcha into the individual words.  For the most of these options, you can mostly leave the defaults unless you feel like experimenting")
    parser.add_argument('filename', help='The spectrogram (from sox) of the audio file to split into individual words')
    parser.add_argument('-b', type=int, required=False, default=140, metavar='black_pixel_count_threshold', help='The maximum number of black pixels a column in the spectrogram can have and still be considered part of a word')
    parser.add_argument('-d', required=False, action='store_true', help='Debug mode.  Prints out an extra spectrogram with the axis removed and each word with a white border')
    parser.add_argument('-m', type=int, required=False, default=2, metavar='maximum_column_xdiff', help='The maximum amount to travel in the x direction (at one time) when looking for the word\'s column in the spectrogram')
    parser.add_argument('-w', type=int, required=False, default=170, metavar='minimum_word_height', help='The divide between background noise and words in the spectrogram.  This value does not have to be exact as long as it is definitely above the background noise.')
    parser.add_argument('-v', required=False, action='store_true', help='Verbose mode')
    parser.add_argument('-vv', required=False, action='store_true', help='Very Verbose mode')

    args = parser.parse_args()
    output_level = logging.CRITICAL
    debug_images = False
    if args.vv:
        output_level = logging.DEBUG
        debug_images = True
    elif args.v:
        output_level = logging.INFO

    captcha_splitter = SpectrogramSplitter(black_pixel_count_threshold = args.b, debug = output_level, debug_images = debug_images, 
                                                    maximum_column_xdiff = args.m, minimum_word_height = args.w, filename_prefix = "")
    filename = " ".join(args.filename.split(".")[:-1]) #Remove the file extension
    captcha_splitter.split_captcha(filename)

