#!/usr/bin/python
from optparse import OptionParser
import sys

class Splot():
    """This class will process the ouput of sox ran with the -stats and -freq 
    flags and output the values which will be used in the machine learning."""
    METHOD_SUM = 1
    METHOD_FIRST_SECTION_ONLY = 2
    METHOD_STATS_ONLY = 4
    METHOD_AVERAGE = 8
    METHOD_TINY_STATS = 16
    METHOD_EVERYTHING = 32
    METHOD_EVERYTHING_VARIABLE_LENGTH = 64
    
    MAXIMUM_WIDTH = 6144
    
    def __init__(self, input_filename = "", method=METHOD_EVERYTHING, lines = ""):
        """This will create an instance of an splot using the method you so desire."""
        if lines == "":
            f = open(filename, 'r')
            self.lines = f.readlines()
            f.close()
        else:
            self.lines = lines
        self.use_method = method

    def extract_data(self, delimiter = ","):
        """This will extract the data from the file and return it as a string using
        whatever method was specified in the constructor"""
        output = ""
        if self.use_method & Splot.METHOD_SUM != 0:
            output += self._extract_sum()
        if self.use_method & Splot.METHOD_FIRST_SECTION_ONLY != 0:
            output += self._extract_first_section()
        if self.use_method & Splot.METHOD_STATS_ONLY != 0:
            output += self._extract_stats()
        if self.use_method & Splot.METHOD_AVERAGE != 0:
            output += self._extract_average()
        if self.use_method & Splot.METHOD_TINY_STATS != 0:
            output += self._extract_tiny_stats()
        if self.use_method & Splot.METHOD_EVERYTHING != 0:
            output += self._extract_everything()
        if self.use_method & Splot.METHOD_EVERYTHING_VARIABLE_LENGTH != 0:
            output += self._extract_everything(False)
        output = output.replace('\n', '')[:-1] # strip the last comma
        if delimiter != ",":
            output = output.replace(",",delimiter)
        return output
    
    def _pull_raw_data(self, max_items=2048):
        """This will pull the data from the file and will sum up the values in
        each section for every frequency point.  This will give us a picture of
        the entire file and puts emphasis on sounds that happen multiple times.
        It also stores the number of frames processes so we can determine the
        averages for each frequency point."""
        data = [] #Use a list of tuples rather than a frequency based dictionary as it is much faster
        good_lines = 0
        for line in self.lines:
            values = line.split('  ')
            # we only want the lines with the frequencies, not the stats at the end of the file
            if len(values) != 2:
                continue
            frequency, amplitude = values[0], values[1]
            if good_lines < max_items: 
                data.append([frequency, 1, float(amplitude)])
            else:
                index = good_lines % max_items # Sox gives us max_items frequency lines.  So we can safely just index on that
                data[index][1] += 1
                data[index][2] += float(amplitude)
            good_lines += 1
        return data
    
    def _extract_average(self):
        """This will extract the data from the file and will sum up the values in
        each section for every frequency point and divide by the number of 
        sections.  This gives us the average for each frequency over the course
        of the entire audio file."""
        out = ""
        data = self._pull_raw_data()
        for [freq, count, amplitude_sum] in data:
            out += str(amplitude_sum / count) + ","
        return out
    
    def _extract_sum(self):
        """This will extract the data from the file and will sum up the values in
        each section for every frequency point.  This will give us a picture of
        the entire file and puts emphasis on sounds that happen multiple times."""
        out = ""
        data = self._pull_raw_data() 
        for [freq, count, amplitude_sum] in data:
            out += str(amplitude_sum) + ","
        return out
    
    def _extract_everything(self, pad_with_zeros=True):
        """This will extract just the data points, but it will do it for all
        available frames.  If the number of frames is greater than MAXIMUM_WIDTH
        it will be truncated, however this should never happen (in theory)."""
        data = []
        for line in self.lines:
            values = line.strip().split('  ')
            # we only want the lines with the frequencies, not the stats at the end of the file
            if len(values) != 2:
                continue
            data.append(str(values[1]))
            if len(data) >= Splot.MAXIMUM_WIDTH:
                break # just in case there's an unexpectedly long one, we'll truncate at MAXIMUM_WIDTH
        
        if pad_with_zeros:
            while len(data) < Splot.MAXIMUM_WIDTH:
                data.append('0') # pad with zeros so they're all the same
        return ",".join(data) + ","
    
    def _extract_first_section(self):
        out = ""
        for line in self.lines[0:2048]:
            if out == "":
                out = line.split('  ')[1]
            else:
                out += "," + line.split('  ')[1]
        return out
    
    def _extract_stats(self):
        out = ""
        for line in self.lines:
            values = line.split('  ')
            if len(values) == 2:
                continue
            values = line.split(' ')
            if values[0] == "Minimum" and values[1] == "delta:":
                continue # this is always zero
            out += line.split(' ')[-1].strip() + ","
        return out

    def _extract_tiny_stats(self):
        out = ""
        for line in self.lines:
            values = line.split('  ')
            if len(values) == 2:
                continue
            values = line.split(' ')
            if values[0] == "Length":
                out += line.split(' ')[-1].strip() + ","
        return out

if __name__ == "__main__":
    parser = OptionParser(usage="%prog <sox_stats_filename> <bitmask>")
    (options, args) = parser.parse_args()
    
    if len(args) < 1:
        parser.print_help()
        print "ERROR: Filename not specified"
        sys.exit(-1)
    
    filename = args[0]
    if len(args) < 2:
        s = Splot(filename)
    else:
        s = Splot(filename, int(args[1]))

    print s.extract_data()

